#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <sys/wait.h>
#include <limits.h>
#include <ctype.h>
#include "parse_escape.h"
#include "replace.h"

#ifndef ARG_MAX
#define ARG_MAX 10000
#endif

static int args;
static int args_added;
static char *cmd[ARG_MAX + 1];

static char *I_flag;
static char t_flag;
static char r_flag;
static char nl_flag;
static int n_flag;
static size_t s_flag;
static char *d_flag;
static char i_flag;

enum {
	NORMAL,
	CEOF,
	R_FLAG,
	ERROR,
	STOP
};

enum {
	NONE,
	QUOTE
};

static void clear_cmd(void) {
	for (int i = 0; i < args; i++) {
		if (cmd[i] != NULL) {
			free(cmd[i]);
			cmd[i] = NULL;
		}
	}

	args = 0;
	args_added = 0;
}

static int add_arg(const char *str, size_t chars, const char flag) {
	if (!flag)
		args_added++;

	int ret = NORMAL;
	if (args >= ARG_MAX)
		return ERROR;

	else if (n_flag > 0 && args_added == n_flag && I_flag == NULL)
		ret = STOP;

	else if (s_flag > 0 && chars >= s_flag)
		ret = STOP;

	if (!flag && I_flag) {
		for (int i = 0; i < args; i++)
			if (strstr(cmd[i], I_flag)) {
				char *ptr = NULL;
CONTINUE:
				/* Ptr will be freed in the clear_cmd() */
				ptr = mu_replace("xargs", cmd[i], str, I_flag);
				free(cmd[i]);

				if (ptr == NULL)
					return ERROR;

				cmd[i] = ptr;
				if (strstr(cmd[i], I_flag))
					goto CONTINUE;
			}

		return STOP;
	}

	cmd[args] = strdup(str);
	args++;

	return ret;
}

static int is_correct(char c) {
	if (nl_flag)
		return (c == '\0');

	else if (d_flag) {
		for (size_t i = 0; i < strlen(d_flag); i++) {
			if (c == d_flag[i])
				return 1;
		}

		return 0;
	}

	else
		return isspace(c);
}

static int xargs(void) {
	size_t arg_size = 0;
	char *arg = malloc(1);
	if (arg == NULL) {
		fprintf(stderr, "xargs: malloc failed\n");
		return ERROR;
	}

	size_t index = 0;
	int flag = NONE;
	int ret = NORMAL;
	size_t chars = 0;
	int args_passed = 0;

	while (1) {
		int c = getchar();
		if (c == EOF) {
			if (flag == QUOTE) {
				fprintf(stderr, "xargs: unterminated quote\n");
				ret = ERROR;
			}

			ret = CEOF;
			break;
		}

		if ((i_flag || flag == NONE) && is_correct(c) && index > 0) {
			arg[index] = '\0';

			int r = add_arg(arg, chars, 0);
			if (r == ERROR) {
				ret = ERROR;
				break;
			}

			else if (r == STOP) {
				free(arg);
				return NORMAL;
			}

			index = 0;
			arg_size = 0;
			args_passed++;
			free(arg);

			arg = malloc(1);
			if (arg == NULL) {
				fprintf(stderr, "xargs: malloc failed\n");
				return ERROR;
			}
		}

		else {
			if (!i_flag && (c == '"' || c == '\'')) {
				if (flag == QUOTE)
					flag = NONE;

				else
					flag = QUOTE;

				continue;
			}

			else if (is_correct(c) && flag == NONE)
				continue;

			char *tmp = realloc(arg, (++arg_size) + 1);
			if (tmp == NULL) {
				free(arg);

				fprintf(stderr, "xargs: realloc failed\n");
				return ERROR;
			}

			arg = tmp;
			arg[index] = (char)c;

			index++;
			chars++;
		}
	}

	if (args_passed == 0 && ret != ERROR)
		ret = R_FLAG;

	free(arg);
	return ret;
}

static int spawn(void) {
	if (t_flag) {
		for (int i = 0; i < args; i++)
			fprintf(stderr, "%s ", cmd[i]);

		fputc('\n', stderr);
	}

	pid_t pid;
	if ((pid = fork()) == 0) {
		execvp(cmd[0], cmd);
		fprintf(stderr, "xargs: exec: %s\n", strerror(errno));
		exit(1);
	}

	int status = 0;
	waitpid(pid, &status, 0);

	if (status == 255)
		status = 124;

	else if (status >= 0x180)
		status = 125;

	return status;
}

int main(int argc, char **argv) {
	/* For -s flag */
	char *p = NULL;
	int ret = 0;

	int opt;
	while ((opt = getopt(argc, argv, "tn:s:rP:0I:d:i")) != -1) {
		switch (opt) {
			case 't':
				t_flag = 1;
				break;

			case 'n':
				n_flag = atoi(optarg);
				if (n_flag <= 0) {
					fprintf(stderr, "xargs: -n: invalid number: %s\n", optarg);
					ret = 1;
				}

				break;

			case 's':
				s_flag = strtoul(optarg, &p, 0);
				if (s_flag <= 0 || *p) {
					fprintf(stderr, "xargs: -s: invalid number: %zu%s\n", s_flag, (p) ? p : "");
					ret = 1;
				}

				break;

			case 'r':
				r_flag = 1;
				break;

			case '0':
				nl_flag = 1;
				break;

			case 'I':
				I_flag = optarg;
				break;

			case 'd':
				d_flag = mu_parse_escape("xargs", optarg);
				if (d_flag == NULL)
					return 1;

				break;

			case 'i':
				i_flag = 1;
				break;

			default:
				puts("xargs [tnsrP0Idti] [cmd [arg1] [arg2...]\n\t-t Print the command before start\n\t-n Pass no more than N args\n\t-r Don't run command if input is empty\n\t-s Pass command line of no more than N bytes\n\t-0 NUL terminated input\n\t-I STR Replace STR within PROG ARGS with input line\n\t-d CHR Use char as delimeter\n\t-i Ignore quotes");
				return 0;
		}
	}

	if (ret)
		goto CLEAN_UP;

	argv += optind;
	argc -= optind;

	while (1) {
		/* Arg */
		if (argc) {
			for (int i = 0; i < argc; i++)
				if (add_arg(argv[i], 0, 1))
					break;
		}

		else
			add_arg("echo", 0, 1);

		int stdin_stat = xargs();
		if (stdin_stat == ERROR)
			ret = 1;

		/* Check NULL */
		for (int i = 0; i < args; i++) {
			if (cmd[i] == NULL) {
				fprintf(stderr, "xargs: strdup failed\n");
				clear_cmd();

				ret = 1;
				goto CLEAN_UP;
			}
		}

		/* Run */
		if (stdin_stat == NORMAL || stdin_stat == CEOF)
			ret = spawn();

		else if (stdin_stat == R_FLAG && !r_flag && I_flag == NULL)
			ret = spawn();

		clear_cmd();
		if (stdin_stat)
			break;
	}

CLEAN_UP:
	if (d_flag)
		free(d_flag);

	return ret;
}
